@@ -2,8 +2,8 @@
  * dpkg - main program for package management
  * main.c - main program
  *
- * Copyright © 1994,1995 Ian Jackson <ian@chiark.greenend.org.uk>
- * Copyright © 2006-2012 Guillem Jover <guillem@debian.org>
+ * Copyright © 1994,1995 Ian Jackson <ijackson@chiark.greenend.org.uk>
+ * Copyright © 2006-2016 Guillem Jover <guillem@debian.org>
  * Copyright © 2010 Canonical Ltd.
  *   written by Martin Pitt <martin.pitt@canonical.com>
  *
@@ -18,14 +18,13 @@
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
- * along with this program.  If not, see <http://www.gnu.org/licenses/>.
+ * along with this program.  If not, see <https://www.gnu.org/licenses/>.
  */
 
 #include <config.h>
 #include <compat.h>
 
 #include <sys/types.h>
-#include <sys/stat.h>
 #include <sys/wait.h>
 
 #include <errno.h>
@@ -33,7 +32,6 @@
 #if HAVE_LOCALE_H
 #include <locale.h>
 #endif
-#include <ctype.h>
 #include <string.h>
 #include <fcntl.h>
 #include <dirent.h>
@@ -44,47 +42,39 @@
 
 #include <dpkg/macros.h>
 #include <dpkg/i18n.h>
+#include <dpkg/c-ctype.h>
 #include <dpkg/dpkg.h>
 #include <dpkg/dpkg-db.h>
 #include <dpkg/arch.h>
 #include <dpkg/subproc.h>
 #include <dpkg/command.h>
-#include <dpkg/pkg-spec.h>
+#include <dpkg/pager.h>
 #include <dpkg/options.h>
+#include <dpkg/db-fsys.h>
 
 #include "main.h"
-#include "filesdb.h"
 #include "filters.h"
 
 static void DPKG_ATTR_NORET
 printversion(const struct cmdinfo *ci, const char *value)
 {
-  printf(_("Debian `%s' package management program version %s.\n"),
-         DPKG, DPKG_VERSION_ARCH);
-  printf(_(
+  if (f_robot) {
+    printf("%s", PACKAGE_VERSION);
+  } else {
+    printf(_("Debian '%s' package management program version %s.\n"),
+           DPKG, PACKAGE_RELEASE);
+    printf(_(
 "This is free software; see the GNU General Public License version 2 or\n"
 "later for copying conditions. There is NO warranty.\n"));
+  }
 
   m_output(stdout, _("<standard output>"));
 
   exit(0);
 }
 
-/* This function, and the config hook for it, can be removed in 14.10 */
-static void
-printfawarn()
-{
-  fprintf(stderr, _(
-"Unknown configuration key `foreign-architecture' found in your `%s'\n"
-"configuration files.  This warning will become a hard error at a later\n"
-"date, so please remove the offending configuration options and replace\n"
-"them with `%s --add-architecture' invocations at the command line.\n\n"),
-  DPKG, DPKG);
-}
-
 /*
  * FIXME: Options that need fixing:
- * dpkg --yet-to-unpack
  * dpkg --command-fd
  */
 
@@ -92,35 +82,40 @@ static void DPKG_ATTR_NORET
 usage(const struct cmdinfo *ci, const char *value)
 {
   printf(_(
-"Usage: %s [<option> ...] <command>\n"
+"Usage: %s [<option>...] <command>\n"
 "\n"), DPKG);
 
   printf(_(
 "Commands:\n"
-"  -i|--install       <.deb file name> ... | -R|--recursive <directory> ...\n"
-"  --unpack           <.deb file name> ... | -R|--recursive <directory> ...\n"
-"  -A|--record-avail  <.deb file name> ... | -R|--recursive <directory> ...\n"
-"  --configure        <package> ... | -a|--pending\n"
-"  --triggers-only    <package> ... | -a|--pending\n"
-"  -r|--remove        <package> ... | -a|--pending\n"
-"  -P|--purge         <package> ... | -a|--pending\n"
-"  --get-selections [<pattern> ...] Get list of selections to stdout.\n"
+"  -i|--install       <.deb file name>... | -R|--recursive <directory>...\n"
+"  --unpack           <.deb file name>... | -R|--recursive <directory>...\n"
+"  -A|--record-avail  <.deb file name>... | -R|--recursive <directory>...\n"
+"  --configure        <package>... | -a|--pending\n"
+"  --triggers-only    <package>... | -a|--pending\n"
+"  -r|--remove        <package>... | -a|--pending\n"
+"  -P|--purge         <package>... | -a|--pending\n"
+"  -V|--verify [<package>...]       Verify the integrity of package(s).\n"
+"  --get-selections [<pattern>...]  Get list of selections to stdout.\n"
 "  --set-selections                 Set package selections from stdin.\n"
 "  --clear-selections               Deselect every non-essential package.\n"
-"  --update-avail <Packages-file>   Replace available packages info.\n"
-"  --merge-avail <Packages-file>    Merge with info from file.\n"
+"  --update-avail [<Packages-file>] Replace available packages info.\n"
+"  --merge-avail [<Packages-file>]  Merge with info from file.\n"
 "  --clear-avail                    Erase existing available info.\n"
 "  --forget-old-unavail             Forget uninstalled unavailable pkgs.\n"
-"  -s|--status <package> ...        Display package status details.\n"
-"  -p|--print-avail <package> ...   Display available version details.\n"
-"  -L|--listfiles <package> ...     List files `owned' by package(s).\n"
-"  -l|--list [<pattern> ...]        List packages concisely.\n"
-"  -S|--search <pattern> ...        Find package(s) owning file(s).\n"
-"  -C|--audit                       Check for broken package(s).\n"
+"  -s|--status [<package>...]       Display package status details.\n"
+"  -p|--print-avail [<package>...]  Display available version details.\n"
+"  -L|--listfiles <package>...      List files 'owned' by package(s).\n"
+"  -l|--list [<pattern>...]         List packages concisely.\n"
+"  -S|--search <pattern>...         Find package(s) owning file(s).\n"
+"  -C|--audit [<package>...]        Check for broken package(s).\n"
+"  --yet-to-unpack                  Print packages selected for installation.\n"
+"  --predep-package                 Print pre-dependencies to unpack.\n"
 "  --add-architecture <arch>        Add <arch> to the list of architectures.\n"
 "  --remove-architecture <arch>     Remove <arch> from the list of architectures.\n"
 "  --print-architecture             Print dpkg architecture.\n"
 "  --print-foreign-architectures    Print allowed foreign architectures.\n"
+"  --assert-<feature>               Assert support for the specified feature.\n"
+"  --validate-<thing> <string>      Validate a <thing>'s <string>.\n"
 "  --compare-versions <a> <op> <b>  Compare version numbers - see below.\n"
 "  --force-help                     Show help on forcing.\n"
 "  -Dh|--debug=help                 Show help on debugging.\n"
@@ -132,21 +127,27 @@ usage(const struct cmdinfo *ci, const char *value)
 "\n"));
 
   printf(_(
-"Use dpkg -b|--build|-c|--contents|-e|--control|-I|--info|-f|--field|\n"
-" -x|--extract|-X|--vextract|--fsys-tarfile  on archives (type %s --help).\n"
-"\n"), BACKEND);
+"Assertable features: support-predepends, working-epoch, long-filenames,\n"
+"  multi-conrep, multi-arch, versioned-provides, protected-field.\n"
+"\n"));
 
   printf(_(
-"For internal use: dpkg --assert-support-predepends | --predep-package |\n"
-"  --assert-working-epoch | --assert-long-filenames | --assert-multi-conrep |\n"
-"  --assert-multi-arch.\n"
+"Validatable things: pkgname, archname, trigname, version.\n"
 "\n"));
 
+  printf(_(
+"Use dpkg with -b, --build, -c, --contents, -e, --control, -I, --info,\n"
+"  -f, --field, -x, --extract, -X, --vextract, --ctrl-tarfile, --fsys-tarfile\n"
+"on archives (type %s --help).\n"
+"\n"), BACKEND);
+
   printf(_(
 "Options:\n"
 "  --admindir=<directory>     Use <directory> instead of %s.\n"
 "  --root=<directory>         Install on a different root directory.\n"
 "  --instdir=<directory>      Change installation dir without changing admin dir.\n"
+"  --pre-invoke=<command>     Set a pre-invoke hook.\n"
+"  --post-invoke=<command>    Set a post-invoke hook.\n"
 "  --path-exclude=<pattern>   Do not install paths which match a shell pattern.\n"
 "  --path-include=<pattern>   Re-include a pattern after a previous exclusion.\n"
 "  -O|--selected-only         Skip packages not selected for install/upgrade.\n"
@@ -154,6 +155,8 @@ usage(const struct cmdinfo *ci, const char *value)
 "  -G|--refuse-downgrade      Skip packages with earlier version than installed.\n"
 "  -B|--auto-deconfigure      Install even if it would break some other package.\n"
 "  --[no-]triggers            Skip or force consequential trigger processing.\n"
+"  --verify-format=<format>   Verify output format (supported: 'rpm').\n"
+"  --no-pager                 Disables the use of any pager.\n"
 "  --no-debsig                Do not try to verify package signatures.\n"
 "  --no-act|--dry-run|--simulate\n"
 "                             Just say what we would do - don't do it.\n"
@@ -161,12 +164,13 @@ usage(const struct cmdinfo *ci, const char *value)
 "  --status-fd <n>            Send status change updates to file descriptor <n>.\n"
 "  --status-logger=<command>  Send status change updates to <command>'s stdin.\n"
 "  --log=<filename>           Log status changes and actions to <filename>.\n"
-"  --ignore-depends=<package>,...\n"
+"  --ignore-depends=<package>[,...]\n"
 "                             Ignore dependencies involving <package>.\n"
-"  --force-...                Override problems (see --force-help).\n"
-"  --no-force-...|--refuse-...\n"
-"                             Stop when problems encountered.\n"
+"  --force-<thing>[,...]      Override problems (see --force-help).\n"
+"  --no-force-<thing>[,...]   Stop when problems encountered.\n"
+"  --refuse-<thing>[,...]     Ditto.\n"
 "  --abort-after <n>          Abort after encountering <n> errors.\n"
+"  --robot                    Use machine-readable output on some commands.\n"
 "\n"), ADMINDIR);
 
   printf(_(
@@ -177,7 +181,7 @@ usage(const struct cmdinfo *ci, const char *value)
 "\n"));
 
   printf(_(
-"Use `dselect' or `aptitude' for user-friendly package management.\n"));
+"Use 'apt' or 'aptitude' for user-friendly package management.\n"));
 
   m_output(stdout, _("<standard output>"));
 
@@ -186,108 +190,23 @@ usage(const struct cmdinfo *ci, const char *value)
 
 static const char printforhelp[] = N_(
 "Type dpkg --help for help about installing and deinstalling packages [*];\n"
-"Use `dselect' or `aptitude' for user-friendly package management;\n"
+"Use 'apt' or 'aptitude' for user-friendly package management;\n"
 "Type dpkg -Dhelp for a list of dpkg debug flag values;\n"
 "Type dpkg --force-help for a list of forcing options;\n"
 "Type dpkg-deb --help for help about manipulating *.deb files;\n"
 "\n"
-"Options marked [*] produce a lot of output - pipe it through `less' or `more' !");
+"Options marked [*] produce a lot of output - pipe it through 'less' or 'more' !");
 
+int f_robot = 0;
 int f_pending=0, f_recursive=0, f_alsoselect=1, f_skipsame=0, f_noact=0;
 int f_autodeconf=0, f_nodebsig=0;
 int f_triggers = 0;
-int fc_downgrade=1, fc_configureany=0, fc_hold=0, fc_removereinstreq=0, fc_overwrite=0;
-int fc_removeessential=0, fc_conflicts=0, fc_depends=0, fc_dependsversion=0;
-int fc_breaks=0, fc_badpath=0, fc_overwritediverted=0, fc_architecture=0;
-int fc_nonroot=0, fc_overwritedir=0, fc_conff_new=0, fc_conff_miss=0;
-int fc_conff_old=0, fc_conff_def=0;
-int fc_conff_ask = 0;
-int fc_unsafe_io = 0;
-int fc_badverify = 0;
-int fc_badversion = 0;
 
 int errabort = 50;
-static const char *admindir = ADMINDIR;
+static const char *admindir;
 const char *instdir= "";
 struct pkg_list *ignoredependss = NULL;
 
-static const char *
-forcetype_str(char type)
-{
-  switch (type) {
-  case '\0':
-  case ' ':
-    return "   ";
-  case '*':
-    return "[*]";
-  case '!':
-    return "[!]";
-  default:
-    internerr("unknown force type '%c'", type);
-  }
-}
-
-static const struct forceinfo {
-  const char *name;
-  int *opt;
-  char type;
-  const char *desc;
-} forceinfos[]= {
-  { "all",                 NULL,
-    '!', N_("Set all force options")},
-  { "downgrade",           &fc_downgrade,
-    '*', N_("Replace a package with a lower version") },
-  { "configure-any",       &fc_configureany,
-    ' ', N_("Configure any package which may help this one") },
-  { "hold",                &fc_hold,
-    ' ', N_("Process incidental packages even when on hold") },
-  { "not-root",            &fc_nonroot,
-    ' ', N_("Try to (de)install things even when not root") },
-  { "bad-path",            &fc_badpath,
-    ' ', N_("PATH is missing important programs, problems likely") },
-  { "bad-verify",          &fc_badverify,
-    ' ', N_("Install a package even if it fails authenticity check") },
-  { "bad-version",         &fc_badversion,
-    ' ', N_("Process even packages with wrong versions") },
-  { "overwrite",           &fc_overwrite,
-    ' ', N_("Overwrite a file from one package with another") },
-  { "overwrite-diverted",  &fc_overwritediverted,
-    ' ', N_("Overwrite a diverted file with an undiverted version") },
-  { "overwrite-dir",       &fc_overwritedir,
-    '!', N_("Overwrite one package's directory with another's file") },
-  { "unsafe-io",           &fc_unsafe_io,
-    '!', N_("Do not perform safe I/O operations when unpacking") },
-  { "confnew",             &fc_conff_new,
-    '!', N_("Always use the new config files, don't prompt") },
-  { "confold",             &fc_conff_old,
-    '!', N_("Always use the old config files, don't prompt") },
-  /* XXX: Handle automatically line wrappings. */
-  { "confdef",             &fc_conff_def,
-    '!', N_("Use the default option for new config files if one\n"
-"                         is available, don't prompt. If no default can be found,\n"
-"                         you will be prompted unless one of the confold or\n"
-"                         confnew options is also given") },
-  { "confmiss",            &fc_conff_miss,
-    '!', N_("Always install missing config files") },
-  { "confask",             &fc_conff_ask,
-    '!', N_("Offer to replace config files with no new versions") },
-  { "architecture",        &fc_architecture,
-    '!', N_("Process even packages with wrong or no architecture") },
-  { "breaks",              &fc_breaks,
-    '!', N_("Install even if it would break another package") },
-  { "conflicts",           &fc_conflicts,
-    '!', N_("Allow installation of conflicting packages") },
-  { "depends",             &fc_depends,
-    '!', N_("Turn all dependency problems into warnings") },
-  { "depends-version",     &fc_dependsversion,
-    '!', N_("Turn dependency version problems into warnings") },
-  { "remove-reinstreq",    &fc_removereinstreq,
-    '!', N_("Remove packages which require installation") },
-  { "remove-essential",    &fc_removeessential,
-    '!', N_("Remove an essential package") },
-  { NULL }
-};
-
 #define DBG_DEF(n, d) \
   { .flag = dbg_##n, .name = #n, .desc = d }
 
@@ -312,7 +231,9 @@ static const struct debuginfo {
   { 0, NULL, NULL }
 };
 
-static void setdebug(const struct cmdinfo *cpi, const char *value) {
+static void
+set_debug(const struct cmdinfo *cpi, const char *value)
+{
   char *endp;
   long mask;
   const struct debuginfo *dip;
@@ -342,19 +263,43 @@ static void setdebug(const struct cmdinfo *cpi, const char *value) {
 }
 
 static void
-setfilter(const struct cmdinfo *cip, const char *value)
+set_no_pager(const struct cmdinfo *ci, const char *value)
+{
+  pager_enable(false);
+
+  /* Let's communicate this to our backends. */
+  setenv("DPKG_PAGER", "cat", 1);
+}
+
+static void
+set_filter(const struct cmdinfo *cip, const char *value)
 {
   filter_add(value, cip->arg_int);
 }
 
-static void setroot(const struct cmdinfo *cip, const char *value) {
-  char *p;
-  instdir= value;
-  m_asprintf(&p, "%s%s", value, ADMINDIR);
-  admindir= p;
+static void
+set_verify_format(const struct cmdinfo *cip, const char *value)
+{
+  if (!verify_set_output(value))
+    badusage(_("unknown verify output format '%s'"), value);
 }
 
-static void ignoredepends(const struct cmdinfo *cip, const char *value) {
+static void
+set_instdir(const struct cmdinfo *cip, const char *value)
+{
+  instdir = dpkg_fsys_set_dir(value);
+}
+
+static void
+set_root(const struct cmdinfo *cip, const char *value)
+{
+  instdir = dpkg_fsys_set_dir(value);
+  admindir = dpkg_fsys_get_path(ADMINDIR);
+}
+
+static void
+set_ignore_depends(const struct cmdinfo *cip, const char *value)
+{
   char *copy, *p;
 
   copy= m_malloc(strlen(value)+2);
@@ -369,14 +314,9 @@ static void ignoredepends(const struct cmdinfo *cip, const char *value) {
   }
   p= copy;
   while (*p) {
-    struct dpkg_error err;
     struct pkginfo *pkg;
 
-    pkg = pkg_spec_parse_pkg(p, &err);
-    if (pkg == NULL)
-      ohshit(_("--%s needs a valid package name but '%.250s' is not: %s"),
-              cip->olong, p, err.str);
-
+    pkg = dpkg_options_parse_pkgname(cip, p);
     pkg_list_prepend(&ignoredependss, pkg);
 
     p+= strlen(p)+1;
@@ -385,25 +325,18 @@ static void ignoredepends(const struct cmdinfo *cip, const char *value) {
   free(copy);
 }
 
-static void setinteger(const struct cmdinfo *cip, const char *value) {
-  long v;
-  char *ep;
-
-  errno = 0;
-  v = strtol(value, &ep, 0);
-  if (value == ep || *ep || v < 0 || v > INT_MAX || errno != 0)
-    badusage(_("invalid integer for --%s: `%.250s'"),cip->olong,value);
-  *cip->iassignto= v;
+static void
+set_integer(const struct cmdinfo *cip, const char *value)
+{
+  *cip->iassignto = dpkg_options_parse_arg_int(cip, value);
 }
 
-static void setpipe(const struct cmdinfo *cip, const char *value) {
+static void
+set_pipe(const struct cmdinfo *cip, const char *value)
+{
   long v;
-  char *ep;
 
-  errno = 0;
-  v = strtol(value, &ep, 0);
-  if (value == ep || *ep || v < 0 || v > INT_MAX || errno != 0)
-    badusage(_("invalid integer for --%s: `%.250s'"),cip->olong,value);
+  v = dpkg_options_parse_arg_int(cip, value);
 
   statusfd_add(v);
 }
@@ -418,42 +351,50 @@ is_invoke_action(enum action action)
   case act_triggers:
   case act_remove:
   case act_purge:
+  case act_arch_add:
+  case act_arch_remove:
     return true;
   default:
     return false;
   }
 }
 
-struct invoke_hook *pre_invoke_hooks = NULL;
-struct invoke_hook **pre_invoke_hooks_tail = &pre_invoke_hooks;
-struct invoke_hook *post_invoke_hooks = NULL;
-struct invoke_hook **post_invoke_hooks_tail = &post_invoke_hooks;
-struct invoke_hook *status_loggers = NULL;
-struct invoke_hook **status_loggers_tail = &status_loggers;
+static struct invoke_list pre_invoke_hooks = {
+  .head = NULL,
+  .tail = &pre_invoke_hooks.head,
+};
+static struct invoke_list post_invoke_hooks = {
+  .head = NULL,
+  .tail = &post_invoke_hooks.head,
+};
+static struct invoke_list status_loggers = {
+  .head = NULL,
+  .tail = &status_loggers.head,
+};
 
 static void
 set_invoke_hook(const struct cmdinfo *cip, const char *value)
 {
-  struct invoke_hook ***hook_tail = cip->arg_ptr;
+  struct invoke_list *hook_list = cip->arg_ptr;
   struct invoke_hook *hook_new;
 
-  hook_new = nfmalloc(sizeof(struct invoke_hook));
-  hook_new->command = nfstrsave(value);
+  hook_new = m_malloc(sizeof(*hook_new));
+  hook_new->command = m_strdup(value);
   hook_new->next = NULL;
 
   /* Add the new hook at the tail of the list to preserve the order. */
-  **hook_tail = hook_new;
-  *hook_tail = &hook_new->next;
+  *hook_list->tail = hook_new;
+  hook_list->tail = &hook_new->next;
 }
 
 static void
-run_invoke_hooks(const char *action, struct invoke_hook *hook_head)
+run_invoke_hooks(const char *action, struct invoke_list *hook_list)
 {
   struct invoke_hook *hook;
 
   setenv("DPKG_HOOK_ACTION", action, 1);
 
-  for (hook = hook_head; hook; hook = hook->next) {
+  for (hook = hook_list->head; hook; hook = hook->next) {
     int status;
 
     /* XXX: As an optimization, use exec instead if no shell metachar are
@@ -468,30 +409,51 @@ run_invoke_hooks(const char *action, struct invoke_hook *hook_head)
 }
 
 static void
-run_status_loggers(struct invoke_hook *hook_head)
+free_invoke_hooks(struct invoke_list *hook_list)
+{
+  struct invoke_hook *hook, *hook_next;
+
+  for (hook = hook_list->head; hook; hook = hook_next) {
+    hook_next = hook->next;
+    free(hook->command);
+    free(hook);
+  }
+}
+
+static int
+run_logger(struct invoke_hook *hook, const char *name)
+{
+  pid_t pid;
+  int p[2];
+
+  m_pipe(p);
+
+  pid = subproc_fork();
+  if (pid == 0) {
+    /* Setup stdin and stdout. */
+    m_dup2(p[0], 0);
+    close(1);
+
+    close(p[0]);
+    close(p[1]);
+
+    command_shell(hook->command, name);
+  }
+  close(p[0]);
+
+  return p[1];
+}
+
+static void
+run_status_loggers(struct invoke_list *hook_list)
 {
   struct invoke_hook *hook;
 
-  for (hook = hook_head; hook; hook = hook->next) {
-    pid_t pid;
-    int p[2];
+  for (hook = hook_list->head; hook; hook = hook->next) {
+    int fd;
 
-    m_pipe(p);
-
-    pid = subproc_fork();
-    if (pid == 0) {
-      /* Setup stdin and stdout. */
-      m_dup2(p[0], 0);
-      close(1);
-
-      close(p[0]);
-      close(p[1]);
-
-      command_shell(hook->command, _("status logger"));
-    }
-    close(p[0]);
-
-    statusfd_add(p[1]);
+    fd = run_logger(hook, _("status logger"));
+    statusfd_add(fd);
   }
 }
 
@@ -501,17 +463,17 @@ arch_add(const char *const *argv)
   struct dpkg_arch *arch;
   const char *archname = *argv++;
 
-  if (archname == NULL)
-    badusage(_("--%s takes one argument"), cipaction->olong);
+  if (archname == NULL || *argv)
+    badusage(_("--%s takes exactly one argument"), cipaction->olong);
 
   dpkg_arch_load_list();
 
   arch = dpkg_arch_add(archname);
   switch (arch->type) {
-  case arch_native:
-  case arch_foreign:
+  case DPKG_ARCH_NATIVE:
+  case DPKG_ARCH_FOREIGN:
     break;
-  case arch_illegal:
+  case DPKG_ARCH_ILLEGAL:
     ohshit(_("architecture '%s' is illegal: %s"), archname,
            dpkg_arch_name_is_illegal(archname));
   default:
@@ -528,27 +490,27 @@ arch_remove(const char *const *argv)
 {
   const char *archname = *argv++;
   struct dpkg_arch *arch;
-  struct pkgiterator *iter;
+  struct pkg_hash_iter *iter;
   struct pkginfo *pkg;
 
-  if (archname == NULL)
-    badusage(_("--%s takes one argument"), cipaction->olong);
+  if (archname == NULL || *argv)
+    badusage(_("--%s takes exactly one argument"), cipaction->olong);
 
   modstatdb_open(msdbrw_readonly);
 
   arch = dpkg_arch_find(archname);
-  if (arch->type != arch_foreign) {
+  if (arch->type != DPKG_ARCH_FOREIGN) {
     warning(_("cannot remove non-foreign architecture '%s'"), arch->name);
     return 0;
   }
 
   /* Check if it's safe to remove the architecture from the db. */
-  iter = pkg_db_iter_new();
-  while ((pkg = pkg_db_iter_next_pkg(iter))) {
-    if (pkg->status < stat_halfinstalled)
+  iter = pkg_hash_iter_new();
+  while ((pkg = pkg_hash_iter_next_pkg(iter))) {
+    if (pkg->status < PKG_STAT_HALFINSTALLED)
       continue;
     if (pkg->installed.arch == arch) {
-      if (fc_architecture)
+      if (in_force(FORCE_ARCHITECTURE))
         warning(_("removing architecture '%s' currently in use by database"),
                 arch->name);
       else
@@ -557,7 +519,7 @@ arch_remove(const char *const *argv)
       break;
     }
   }
-  pkg_db_iter_free(iter);
+  pkg_hash_iter_free(iter);
 
   dpkg_arch_unmark(arch);
   dpkg_arch_save_list();
@@ -567,55 +529,6 @@ arch_remove(const char *const *argv)
   return 0;
 }
 
-static void setforce(const struct cmdinfo *cip, const char *value) {
-  const char *comma;
-  size_t l;
-  const struct forceinfo *fip;
-
-  if (strcmp(value, "help") == 0) {
-    printf(_(
-"%s forcing options - control behaviour when problems found:\n"
-"  warn but continue:  --force-<thing>,<thing>,...\n"
-"  stop with error:    --refuse-<thing>,<thing>,... | --no-force-<thing>,...\n"
-" Forcing things:\n"), DPKG);
-
-    for (fip = forceinfos; fip->name; fip++)
-      printf("  %s %-18s %s\n", forcetype_str(fip->type), fip->name,
-             gettext(fip->desc));
-
-    printf(_(
-"\n"
-"WARNING - use of options marked [!] can seriously damage your installation.\n"
-"Forcing options marked [*] are enabled by default.\n"));
-    m_output(stdout, _("<standard output>"));
-    exit(0);
-  }
-
-  for (;;) {
-    comma= strchr(value,',');
-    l = comma ? (size_t)(comma - value) : strlen(value);
-    for (fip=forceinfos; fip->name; fip++)
-      if (strncmp(fip->name, value, l) == 0 && strlen(fip->name) == l)
-        break;
-
-    if (!fip->name) {
-      badusage(_("unknown force/refuse option `%.*s'"),
-               (int)min(l, 250), value);
-    } else if (strcmp(fip->name, "all") == 0) {
-      for (fip = forceinfos; fip->name; fip++)
-        if (fip->opt)
-          *fip->opt = cip->arg_int;
-    } else if (fip->opt) {
-      *fip->opt = cip->arg_int;
-    } else {
-      warning(_("obsolete force/refuse option '%s'\n"), fip->name);
-    }
-
-    if (!comma) break;
-    value= ++comma;
-  }
-}
-
 int execbackend(const char *const *argv) DPKG_ATTR_NORET;
 int commandfd(const char *const *argv);
 
@@ -633,6 +546,7 @@ static const struct cmdinfo cmdinfos[]= {
   ACTION( "remove",                         'r', act_remove,               packages        ),
   ACTION( "purge",                          'P', act_purge,                packages        ),
   ACTION( "triggers-only",                   0,  act_triggers,             packages        ),
+  ACTION( "verify",                         'V', act_verify,               verify          ),
   ACTIONBACKEND( "listfiles",               'L', DPKGQUERY),
   ACTIONBACKEND( "status",                  's', DPKGQUERY),
   ACTION( "get-selections",                  0,  act_getselections,        getselections   ),
@@ -652,32 +566,39 @@ static const struct cmdinfo cmdinfos[]= {
   ACTION( "assert-long-filenames",           0,  act_assertlongfilenames,  assertlongfilenames ),
   ACTION( "assert-multi-conrep",             0,  act_assertmulticonrep,    assertmulticonrep ),
   ACTION( "assert-multi-arch",               0,  act_assertmultiarch,      assertmultiarch ),
+  ACTION( "assert-versioned-provides",       0,  act_assertverprovides,    assertverprovides ),
+  ACTION( "assert-protected-field",          0,  act_assert_protected,     assert_protected ),
   ACTION( "add-architecture",                0,  act_arch_add,             arch_add        ),
   ACTION( "remove-architecture",             0,  act_arch_remove,          arch_remove     ),
   ACTION( "print-architecture",              0,  act_printarch,            printarch   ),
-  ACTION( "print-installation-architecture", 0,  act_printinstarch,        printinstarch  ),
   ACTION( "print-foreign-architectures",     0,  act_printforeignarches,   print_foreign_arches ),
   ACTION( "predep-package",                  0,  act_predeppackage,        predeppackage   ),
+  ACTION( "validate-pkgname",                0,  act_validate_pkgname,     validate_pkgname ),
+  ACTION( "validate-trigname",               0,  act_validate_trigname,    validate_trigname ),
+  ACTION( "validate-archname",               0,  act_validate_archname,    validate_archname ),
+  ACTION( "validate-version",                0,  act_validate_version,     validate_version ),
   ACTION( "compare-versions",                0,  act_cmpversions,          cmpversions     ),
 /*
   ACTION( "command-fd",                   'c', act_commandfd,   commandfd     ),
 */
 
-  { "pre-invoke",        0,   1, NULL,          NULL,      set_invoke_hook, 0, &pre_invoke_hooks_tail },
-  { "post-invoke",       0,   1, NULL,          NULL,      set_invoke_hook, 0, &post_invoke_hooks_tail },
-  { "path-exclude",      0,   1, NULL,          NULL,      setfilter,     0 },
-  { "path-include",      0,   1, NULL,          NULL,      setfilter,     1 },
-  { "status-logger",     0,   1, NULL,          NULL,      set_invoke_hook, 0, &status_loggers_tail },
-  { "status-fd",         0,   1, NULL,          NULL,      setpipe, 0 },
+  { "pre-invoke",        0,   1, NULL,          NULL,      set_invoke_hook, 0, &pre_invoke_hooks },
+  { "post-invoke",       0,   1, NULL,          NULL,      set_invoke_hook, 0, &post_invoke_hooks },
+  { "path-exclude",      0,   1, NULL,          NULL,      set_filter,     0 },
+  { "path-include",      0,   1, NULL,          NULL,      set_filter,     1 },
+  { "verify-format",     0,   1, NULL,          NULL,      set_verify_format },
+  { "status-logger",     0,   1, NULL,          NULL,      set_invoke_hook, 0, &status_loggers },
+  { "status-fd",         0,   1, NULL,          NULL,      set_pipe, 0 },
   { "log",               0,   1, NULL,          &log_file, NULL,    0 },
   { "pending",           'a', 0, &f_pending,    NULL,      NULL,    1 },
   { "recursive",         'R', 0, &f_recursive,  NULL,      NULL,    1 },
   { "no-act",            0,   0, &f_noact,      NULL,      NULL,    1 },
   { "dry-run",           0,   0, &f_noact,      NULL,      NULL,    1 },
   { "simulate",          0,   0, &f_noact,      NULL,      NULL,    1 },
+  { "no-pager",          0,   0, NULL,          NULL,      set_no_pager,  0 },
   { "no-debsig",         0,   0, &f_nodebsig,   NULL,      NULL,    1 },
   /* Alias ('G') for --refuse. */
-  {  NULL,               'G', 0, &fc_downgrade, NULL,      NULL,    0 },
+  {  NULL,               'G', 0, NULL,          NULL,      reset_force_option, FORCE_DOWNGRADE },
   { "selected-only",     'O', 0, &f_alsoselect, NULL,      NULL,    0 },
   { "triggers",           0,  0, &f_triggers,   NULL,      NULL,    1 },
   { "no-triggers",        0,  0, &f_triggers,   NULL,      NULL,   -1 },
@@ -685,19 +606,18 @@ static const struct cmdinfo cmdinfos[]= {
   { "no-also-select",    'N', 0, &f_alsoselect, NULL,      NULL,    0 },
   { "skip-same-version", 'E', 0, &f_skipsame,   NULL,      NULL,    1 },
   { "auto-deconfigure",  'B', 0, &f_autodeconf, NULL,      NULL,    1 },
-  { "root",              0,   1, NULL,          NULL,      setroot,       0 },
-  { "abort-after",       0,   1, &errabort,     NULL,      setinteger,    0 },
+  { "robot",             0,   0, &f_robot,      NULL,      NULL,    1 },
+  { "root",              0,   1, NULL,          NULL,      set_root,      0 },
+  { "abort-after",       0,   1, &errabort,     NULL,      set_integer,   0 },
   { "admindir",          0,   1, NULL,          &admindir, NULL,          0 },
-  { "instdir",           0,   1, NULL,          &instdir,  NULL,          0 },
-  { "ignore-depends",    0,   1, NULL,          NULL,      ignoredepends, 0 },
-  { "force",             0,   2, NULL,          NULL,      setforce,      1 },
-  { "refuse",            0,   2, NULL,          NULL,      setforce,      0 },
-  { "no-force",          0,   2, NULL,          NULL,      setforce,      0 },
-  { "debug",             'D', 1, NULL,          NULL,      setdebug,      0 },
+  { "instdir",           0,   1, NULL,          NULL,      set_instdir,   0 },
+  { "ignore-depends",    0,   1, NULL,          NULL,      set_ignore_depends, 0 },
+  { "force",             0,   2, NULL,          NULL,      set_force_option,   1 },
+  { "refuse",            0,   2, NULL,          NULL,      set_force_option,   0 },
+  { "no-force",          0,   2, NULL,          NULL,      set_force_option,   0 },
+  { "debug",             'D', 1, NULL,          NULL,      set_debug,     0 },
   { "help",              '?', 0, NULL,          NULL,      usage,         0 },
   { "version",           0,   0, NULL,          NULL,      printversion,  0 },
-  /* This config hook, and the function it calls, can be removed in 14.10 */
-  { "foreign-architecture", 0, 1, NULL,         NULL,      printfawarn,   0 },
   ACTIONBACKEND( "build",		'b', BACKEND),
   ACTIONBACKEND( "contents",		'c', BACKEND),
   ACTIONBACKEND( "control",		'e', BACKEND),
@@ -705,6 +625,7 @@ static const struct cmdinfo cmdinfos[]= {
   ACTIONBACKEND( "field",		'f', BACKEND),
   ACTIONBACKEND( "extract",		'x', BACKEND),
   ACTIONBACKEND( "vextract",		'X', BACKEND),
+  ACTIONBACKEND( "ctrl-tarfile",	0,   BACKEND),
   ACTIONBACKEND( "fsys-tarfile",	0,   BACKEND),
   { NULL,                0,   0, NULL,          NULL,      NULL,          0 }
 };
@@ -713,15 +634,12 @@ int
 execbackend(const char *const *argv)
 {
   struct command cmd;
-  char *arg;
 
   command_init(&cmd, cipaction->arg_ptr, NULL);
   command_add_arg(&cmd, cipaction->arg_ptr);
+  command_add_arg(&cmd, str_fmt("--%s", cipaction->olong));
 
-  m_asprintf(&arg, "--%s", cipaction->olong);
-  command_add_arg(&cmd, arg);
-
-  /* Exlicitely separate arguments from options as any user-supplied
+  /* Explicitly separate arguments from options as any user-supplied
    * separator got stripped by the option parser */
   command_add_arg(&cmd, "--");
   command_add_argl(&cmd, (const char **)argv);
@@ -734,7 +652,7 @@ commandfd(const char *const *argv)
 {
   struct varbuf linevb = VARBUF_INIT;
   const char * pipein;
-  const char **newargs = NULL;
+  const char **newargs = NULL, **endargs;
   char *ptr, *endptr;
   FILE *in;
   long infd;
@@ -745,12 +663,11 @@ commandfd(const char *const *argv)
   pipein = *argv++;
   if (pipein == NULL || *argv)
     badusage(_("--%s takes exactly one argument"), cipaction->olong);
-  errno = 0;
-  infd = strtol(pipein, &endptr, 10);
-  if (pipein == endptr || *endptr || infd < 0 || infd > INT_MAX || errno != 0)
-    ohshit(_("invalid integer for --%s: `%.250s'"), cipaction->olong, pipein);
-  if ((in= fdopen(infd, "r")) == NULL)
-    ohshite(_("couldn't open `%i' for stream"), (int) infd);
+
+  infd = dpkg_options_parse_arg_int(cipaction, pipein);
+  in = fdopen(infd, "r");
+  if (in == NULL)
+    ohshite(_("couldn't open '%i' for stream"), (int)infd);
 
   for (;;) {
     bool mode = false;
@@ -759,7 +676,11 @@ commandfd(const char *const *argv)
 
     push_error_context();
 
-    do { c= getc(in); if (c == '\n') lno++; } while (c != EOF && isspace(c));
+    do {
+      c = getc(in);
+      if (c == '\n')
+        lno++;
+    } while (c != EOF && c_isspace(c));
     if (c == EOF) break;
     if (c == '#') {
       do { c= getc(in); if (c == '\n') lno++; } while (c != EOF && c != '\n');
@@ -772,10 +693,11 @@ commandfd(const char *const *argv)
       if (c == '\n') lno++;
 
       /* This isn't fully accurate, but overestimating can't hurt. */
-      if (isspace(c))
+      if (c_isspace(c))
         argc++;
     } while (c != EOF && c != '\n');
-    if (c == EOF) ohshit(_("unexpected eof before end of line %d"),lno);
+    if (c == EOF)
+      ohshit(_("unexpected end of file before end of line %d"), lno);
     if (!argc) continue;
     varbuf_end_str(&linevb);
     newargs = m_realloc(newargs, sizeof(const char *) * (argc + 1));
@@ -791,7 +713,7 @@ commandfd(const char *const *argv)
 	endptr--;
 	skipchar = true;
 	continue;
-      } else if (isspace(*ptr)) {
+      } else if (c_isspace(*ptr)) {
 	if (mode == true) {
 	  *ptr = '\0';
 	  mode = false;
@@ -815,31 +737,39 @@ commandfd(const char *const *argv)
     for(i=1;i<argc;i++)
       if (newargs[i])
         newargs[i] = m_strdup(newargs[i]);
+    endargs = newargs;
 
     setaction(NULL, NULL);
-    myopt((const char *const **)&newargs, cmdinfos, printforhelp);
+    dpkg_options_parse((const char *const **)&endargs, cmdinfos, printforhelp);
     if (!cipaction) badusage(_("need an action option"));
 
-    ret |= cipaction->action(newargs);
+    ret |= cipaction->action(endargs);
+
+    fsys_hash_reset();
 
     pop_error_context(ehflag_normaltidy);
   }
 
+  fclose(in);
+
   return ret;
 }
 
 int main(int argc, const char *const *argv) {
+  char *force_string;
   int ret;
 
-  if (getenv("DPKG_UNTRANSLATED_MESSAGES") == NULL)
-    setlocale(LC_ALL, "");
-  bindtextdomain(PACKAGE, LOCALEDIR);
-  textdomain(PACKAGE);
+  dpkg_locales_init(PACKAGE);
+  dpkg_program_init("dpkg");
+  set_force_default(FORCE_ALL);
+  dpkg_options_load(DPKG, cmdinfos);
+  dpkg_options_parse(&argv, cmdinfos, printforhelp);
 
-  dpkg_set_progname("dpkg");
-  standard_startup();
-  loadcfgfile(DPKG, cmdinfos);
-  myopt(&argv, cmdinfos, printforhelp);
+  /* When running as root, make sure our primary group is also root, so
+   * that files created by maintainer scripts have correct ownership. */
+  if (!in_force(FORCE_NON_ROOT) && getuid() == 0)
+    if (setgid(0) < 0)
+      ohshite(_("cannot set primary group ID to root"));
 
   if (!cipaction) badusage(_("need an action option"));
 
@@ -848,25 +778,31 @@ int main(int argc, const char *const *argv) {
   /* Always set environment, to avoid possible security risks. */
   if (setenv("DPKG_ADMINDIR", admindir, 1) < 0)
     ohshite(_("unable to setenv for subprocesses"));
+  if (setenv("DPKG_ROOT", instdir, 1) < 0)
+    ohshite(_("unable to setenv for subprocesses"));
+  force_string = get_force_string();
+  if (setenv("DPKG_FORCE", force_string, 1) < 0)
+    ohshite(_("unable to setenv for subprocesses"));
+  free(force_string);
 
   if (!f_triggers)
     f_triggers = (cipaction->arg_int == act_triggers && *argv) ? -1 : 1;
 
-  setvbuf(stdout, NULL, _IONBF, 0);
-
   if (is_invoke_action(cipaction->arg_int)) {
-    run_invoke_hooks(cipaction->olong, pre_invoke_hooks);
-    run_status_loggers(status_loggers);
+    run_invoke_hooks(cipaction->olong, &pre_invoke_hooks);
+    run_status_loggers(&status_loggers);
   }
 
-  filesdbinit();
-
   ret = cipaction->action(argv);
 
   if (is_invoke_action(cipaction->arg_int))
-    run_invoke_hooks(cipaction->olong, post_invoke_hooks);
+    run_invoke_hooks(cipaction->olong, &post_invoke_hooks);
 
-  standard_shutdown();
+  free_invoke_hooks(&pre_invoke_hooks);
+  free_invoke_hooks(&post_invoke_hooks);
+
+  dpkg_program_done();
+  dpkg_locales_done();
 
   return reportbroken_retexitstatus(ret);
 }
